from django_filters.filters import ModelChoiceFilter, ModelMultipleChoiceFilter
from django.utils.translation import ugettext as _
from django.db.models import Q
from functools import reduce

from teryt_tree.models import JednostkaAdministracyjna


class AreaFilter(ModelChoiceFilter):
    def __init__(self, *args, **kwargs):
        label = kwargs.pop("label", _("Area"))
        required = kwargs.pop("required", False)
        queryset = kwargs.pop("queryset", JednostkaAdministracyjna.objects.all())
        method = kwargs.pop("method", lambda q, _, v: q.area(v))
        super(ModelChoiceFilter, self).__init__(
            *args,
            label=label,
            required=required,
            queryset=queryset,
            method=method,
            **kwargs
        )

    @staticmethod
    def filter_area(queryset, value, area_field=None):
        """
        Filter by area.

        `area_field` is the field which points to a field of type `JednostkaAdministracyjna`.

        This method is not automatically injected to the queryset.
        It merely serves as a sample implementation of an `area` filter.
        """

        def with_area_field(path):
            if area_field is None:
                return path
            else:
                return "{}__{}".format(area_field, path)

        return queryset.filter(
            **{
                with_area_field("tree_id"): value.tree_id,
                with_area_field("lft__range"): (value.lft, value.rght),
            }
        )


class AreaMultipleFilter(ModelMultipleChoiceFilter):
    """
    Multiple choice filter for JSTs.
    The queryset should implement a `area_in` method.
    """

    def __init__(self, *args, **kwargs):
        label = kwargs.pop("label", _("Area"))
        required = kwargs.pop("required", False)
        queryset = kwargs.pop("queryset", JednostkaAdministracyjna.objects.all())
        method = kwargs.pop("method", lambda q, _, v: q.area_in(v))
        super(ModelMultipleChoiceFilter, self).__init__(
            *args,
            label=label,
            required=required,
            queryset=queryset,
            method=method,
            **kwargs
        )

    @staticmethod
    def filter_area_in(queryset, value, area_field=None):
        """
        Multiselect filter by area.
        If `value` is empty, returns an empty queryset.

        `area_field` is the field which points to a field of type `JednostkaAdministracyjna`.

        This method is not automatically injected to the queryset.
        It merely serves as a sample implementation of an `area_in` filter.
        """

        def with_area_field(path):
            if area_field is None:
                return path
            else:
                return "{}__{}".format(area_field, path)

        if not value:
            return queryset.none()

        return queryset.filter(
            reduce(
                lambda x, y: x | y,
                [
                    Q(
                        **{
                            with_area_field("tree_id"): jst.tree_id,
                            with_area_field("lft__range"): (jst.lft, jst.rght),
                        }
                    )
                    for jst in value
                ],
            )
        )


class VoivodeshipFilter(AreaFilter):
    def __init__(self, *args, **kwargs):
        label = kwargs.pop("label", _("Voivodeship"))
        queryset = kwargs.pop(
            "queryset", JednostkaAdministracyjna.objects.voivodeship().all()
        )
        super().__init__(*args, label=label, queryset=queryset, **kwargs)


class CountyFilter(AreaFilter):
    def __init__(self, *args, **kwargs):
        label = kwargs.pop("label", _("County"))
        queryset = kwargs.pop(
            "queryset", JednostkaAdministracyjna.objects.county().all()
        )
        super().__init__(*args, label=label, queryset=queryset, **kwargs)


class CommunityFilter(AreaFilter):
    def __init__(self, *args, **kwargs):
        label = kwargs.pop("label", _("Community"))
        queryset = kwargs.pop(
            "queryset", JednostkaAdministracyjna.objects.community().all()
        )
        super().__init__(*args, label=label, queryset=queryset, **kwargs)



from django.test import TestCase
from django_filters.filterset import FilterSet

from teryt_tree import factories
from teryt_tree.dal_ext import filters

from ..models import TestModel


class F(FilterSet):
    """
    FilterSet built on top of filters under test.
    Based on https://github.com/carltongibson/django-filter/blob/master/tests/test_filtering.py
    """

    area = filters.AreaFilter()

    class Meta:
        model = TestModel
        fields = ["area"]


class FMulti(FilterSet):
    area = filters.AreaMultipleFilter()

    class Meta:
        model = TestModel
        fields = ["area"]


class AreaFilterTestCase(TestCase):
    def setUp(self):
        self.jst_1 = factories.JednostkaAdministracyjnaFactory()
        self.jst_1_child = factories.JednostkaAdministracyjnaFactory(parent=self.jst_1)
        self.jst_2 = factories.JednostkaAdministracyjnaFactory()
        self.model_1 = TestModel.objects.create(area=self.jst_1)
        self.model_1_child = TestModel.objects.create(area=self.jst_1_child)
        self.model_2 = TestModel.objects.create(area=self.jst_2)

    def assertQsIdsEqual(self, qs, objects):
        return self.assertQuerysetEqual(
            qs, [o.pk for o in objects], lambda o: o.pk, ordered=False
        )

    def test_filter_no_params(self):
        qs = TestModel.objects.all()
        f = F({}, queryset=qs)
        self.assertQsIdsEqual(f.qs, [self.model_1, self.model_1_child, self.model_2])

    def test_filter_single_with_child(self):
        qs = TestModel.objects.all()
        f = F({"area": self.jst_1.pk}, queryset=qs)
        self.assertQsIdsEqual(f.qs, [self.model_1, self.model_1_child])

    def test_filter_unknown_id(self):
        qs = TestModel.objects.all()
        f = F({"area": -1}, queryset=qs)
        self.assertQsIdsEqual(f.qs, [self.model_1, self.model_1_child, self.model_2])


class AreaMultipleFilterTestCase(TestCase):
    def setUp(self):
        self.jst_1 = factories.JednostkaAdministracyjnaFactory()
        self.jst_1_child = factories.JednostkaAdministracyjnaFactory(parent=self.jst_1)
        self.jst_2 = factories.JednostkaAdministracyjnaFactory()
        self.model_1 = TestModel.objects.create(area=self.jst_1)
        self.model_1_child = TestModel.objects.create(area=self.jst_1_child)
        self.model_2 = TestModel.objects.create(area=self.jst_2)

    def assertQsIdsEqual(self, qs, objects):
        return self.assertQuerysetEqual(
            qs, [o.pk for o in objects], lambda o: o.pk, ordered=False
        )

    def test_filter_no_params(self):
        qs = TestModel.objects.all()
        f = FMulti({}, queryset=qs)
        self.assertQsIdsEqual(f.qs, [])

    def test_filter_no_ids(self):
        qs = TestModel.objects.all()
        f = FMulti({"area": []}, queryset=qs)
        self.assertQsIdsEqual(f.qs, [])

    def test_filter_single_with_child(self):
        qs = TestModel.objects.all()
        f = FMulti({"area": [self.jst_1.pk]}, queryset=qs)
        self.assertQsIdsEqual(f.qs, [self.model_1, self.model_1_child])

    def test_filter_multiple(self):
        qs = TestModel.objects.all()
        f = FMulti({"area": [self.jst_1.pk, self.jst_2.pk]}, queryset=qs)
        self.assertQsIdsEqual(f.qs, [self.model_1, self.model_1_child, self.model_2])



from django.db import models

from teryt_tree.models import JednostkaAdministracyjna
from teryt_tree.dal_ext.filters import AreaFilter, AreaMultipleFilter


class TestModelQuerySet(models.QuerySet):
    def area(self, jst):
        return AreaFilter.filter_area(self, jst, "area")

    def area_in(self, jst):
        return AreaMultipleFilter.filter_area_in(self, jst, "area")


class TestModel(models.Model):
    area = models.ForeignKey(JednostkaAdministracyjna, on_delete=models.CASCADE)
    objects = TestModelQuerySet.as_manager()
